#!/usr/bin/env python

"""
tcpdump_watch1.py
by Javantea
April 12, 2008

Watch packets in groups.

For this to work properly in realtime, it will probably (likely) need to be 
threaded for slow networks. It'll work without threading for this version 
though.
"""

from sys import stdin, stdout
from time import time, sleep
import curses

def reset_packet_bucket(bucket):
	output = {}
	for k in bucket.keys():
		output[k] = 0
	#end if
	return output
#end def reset_packet_bucket(bucket)

def time_parse(stime):
	try:
		(h, m, s) = stime.split(':')
	except:
		#print 'error invalid time', stime
		return 0
	#end try
	c_secs = int(h)*3600 + int(m)*60 + float(s)
	return c_secs
#end def time_parse(stime)

def init_curses():
	screen1 = curses.initscr()
	curses.start_color()
	#curses.noecho()
	#curses.cbreak()
	#screen.keypad(1)
	
	curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_WHITE)
	curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)
	curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_RED)
	curses.init_pair(4, curses.COLOR_BLACK, curses.COLOR_YELLOW)
	curses.init_pair(5, curses.COLOR_BLACK, curses.COLOR_GREEN)
	curses.init_pair(6, curses.COLOR_BLACK, curses.COLOR_CYAN)
	screen1.bkgd(' ', curses.color_pair(1))
	return screen1
#end def init_curses()

def draw_buckets(screen, bucket_packet, bucket_size):
	#screen.addstr(0, 0, "TCPdump Watch by Javantea", curses.color_pair(2))
	width = 60
	i = 2
	bucket_keys = bucket_packet.keys()
	bucket_keys.sort()
	for k in bucket_keys:
		pbar_length = min(width-18, bucket_packet[k] / 10)
		packets = str(bucket_packet[k])
		data_rate = int(bucket_size[k]) / (128.0)
		size1 = '%3.2f' % (data_rate)
		bcolor = curses.color_pair(5)
		if data_rate > 300:
			bcolor = curses.color_pair(4)
		elif data_rate > 1000:
			bcolor = curses.color_pair(3)
		elif data_rate > 4000:
			bcolor = curses.color_pair(2)
		#end if
		screen.addstr(i, 0, k + ' ' * (12-len(packets)-len(k)) + packets, curses.color_pair(1))
		screen.addstr(i, 13, size1 + 'kbps' + '*' * (pbar_length-len(size1)), bcolor)
		screen.addstr(' ' * (width-pbar_length-len(size1)-4), curses.color_pair(1))
		#screen.hline(i, 12, '*', pbar_length)
		i += 1
		if i > 12:
			break
		#end if
	#next k
	screen.refresh()
	#sleep(1)
#end def draw_buckets(screen, bucket_packet, bucket_size)

def end_curses(screen):
	curses.nocbreak()
	screen.keypad(0)
	curses.echo()
	
	curses.endwin()
#end def end_curses()

def read_packet_loop(infile, ip=None, screen1=None):
	if screen1:
		screen1.addstr(0, 0, "TCPdump Watch by Javantea", curses.color_pair(6))
	#end if
	
	if ip == None:
		# Automatic ip detection works in the two cases we need it to 
		# (nat and server) without sending a packet.
		from socket import socket, AF_INET, SOCK_STREAM, SOCK_DGRAM
		s = socket(AF_INET, SOCK_DGRAM)
		try:
			s.connect(('216.127.61.180',4569))
		except:
			ip = '127.0.0.1'
		#end try
		(ip, local_udp_port) = s.getsockname()
		s = None
	#end if
	start_time = time()
	time_set = {}
	next_time = 0
	while True:
		o = read_packet_second(infile, ip, next_time)
		if o == None:
			break
		#end if
		(next_time, curr_packet_buckets, curr_size_buckets) = o
		if screen1:
			draw_buckets(screen1, curr_packet_buckets, curr_size_buckets)
		#else:
		#	print 'stats', curr_packet_buckets
		#end if
		#sleep(1)
	#loop
#end def read_packet_loop(infile, ip=None, screen1=None)

def read_packet_second(infile, my_ip=None, next_time = 0):
	"""
	We need an ip to determine input and output. If there's no ip, we assume 
	that everything is out (which isn't always true).
	"""
	# The initial buckets.
	# TODO: IPv6 'IP6 in': 0, 'IP6 out': 0,
	curr_packet_buckets = {'UDP in': 0, 'UDP out': 0, 'UDP': 0, 'TCP in': 0, 'TCP out': 0, 'TCP': 0, 'IP in': 0, 'IP out': 0, 'IP': 0, 'IP6': 0, 'non-IP': 0}
	# copy the buckets.
	curr_size_buckets = dict(curr_packet_buckets)
	while True:
		l2type = 'non-IP'
		l3type = None
		# At least one byte..
		packet_len = 1
		data = infile.readline()
		if data == '':
			break
		#end if
		f = data.split(' ')
		if len(f) < 2:
			#stdout.write(data)
			continue
		#end if
		recv_time = f[0]
		my_direction = ''
		if f[1] == 'IP' or f[1] == 'IP6':
			src_host  = f[2]
			direction = f[3] # is it always > ?
			dst_host  = f[4]
			flags     = f[5] # Syn = S, Ack = A, ...
			if dst_host.startswith(my_ip):
				my_direction = 'in'
			if src_host.startswith(my_ip):
				my_direction = 'out'
			#end if
			packet_len = 20
			l2type = f[1]
			if flags == 'UDP,':
				l3type = 'UDP'
				packet_len = f[7]
				#print 'udp', recv_time, src_host, direction, dst_host, packet_len
			elif flags == 'igmp':
				#l3type = 'IGMP'
				igmp = True
				action = f[6]
				ip = f[7]
				#print 'igmp', recv_time, src_host, direction, dst_host, action, ip
			elif flags == 'VRRPv2,':
				#l3type = 'VRRPv2'
				#print 'VRRPv2', recv_time, src_host, direction, dst_host, f[6:]
				pass
			elif flags == '(frag':
				l3type = 'TCP'
				frag_no = f[6]
				#print 'tcpf', recv_time, src_host, direction, dst_host, frag_no, f[7:]
			else:
				if len(flags) > 1:
					#print 'learn ip flag', f
					continue
				#end if
				l3type = 'TCP'
				data_seqno = f[6]
				#ack = f[6]
				if len(f) < 9:
					win = 'learnw ' + repr(f)
				else:
					win = f[8]
				#end if
				if len(f) < 11:
					mss = 'nomss' #'learnm ' + repr(f)
				else:
					mss = f[10]
				#end if
				#print 'tcp', recv_time, src_host, direction, dst_host, win, mss, f[11:]
			#end if
		elif f[1] == 'arp':
			#l2type = 'ARP'
			arp_type = f[2]
			#print 'arp', recv_time, arp_type, f[3:]
		elif f[1] == 'STP':
			#l2type = 'STP'
			proto_type = f[2]
			#print 'stp', recv_time, proto_type, f[3:]
		elif f[1] == 'CDPv2,':
			#l2type = 'CDPv2'
			proto_type = f[2]
			#print 'stp', recv_time, proto_type, f[3:]
		else:
			if (len(f[0]) > 0) and (f[0][0] == '\t'):
				# this is a part of the previous packet.
				#print data,
				pass
			else:
				pass
				#print 'learn', f
			#end if
			continue
		#end if
		c_secs = time_parse(recv_time)
		if c_secs > next_time:
			#print 'stats', curr_packet_buckets
			#print 'stats', curr_size_buckets
			next_time = c_secs + 1
			return (next_time, curr_packet_buckets, curr_size_buckets)
		#end if
		if my_direction != '':
			l2type = l2type + ' ' + my_direction
		#end if
		curr_packet_buckets[l2type] += 1
		curr_size_buckets[l2type] += int(packet_len)
		if l3type:
			if my_direction != '':
				l3type = l3type + ' ' + my_direction
			#end if
			curr_packet_buckets[l3type] += 1
			curr_size_buckets[l3type] += int(packet_len)
		#end if
	#loop
	return None
#end def read_packet_second(infile, ip=None, next_time=0)

def main():
	"""
	Watch the amount of traffic passing by.
	Usage:
	# Pretty curses view:
	sudo test; sudo /usr/sbin/tcpdump -n -l 2>/dev/null | python tcpdump_watch1.py
	# Without curses:
	sudo /usr/sbin/tcpdump -n -l 2>/dev/null | python tcpdump_watch1.py -c
	"""
	from sys import argv
	# fifo is useful from ssh or some weird thing.
	use_fifo  = '-f' in argv
	# stdin is a bit more stable and probably works with ssh.
	use_stdin  = not use_fifo
	# Curses screen is really awesome. If you see it, you probably like it.
	use_screen = '-c' not in argv
	ip = None
	if '-i' in argv:
		p = argv.index('-i') + 1
		if p < len(argv):
			ip = argv[p]
		#end if
	#end if
	if use_fifo:
		import os
		os.mkfifo('tcpd_watch.pipe')
		infile = file('tcpd_watch.pipe', 'r')
	elif use_stdin:
		infile = stdin
	else:
		#print "Turn on fifo or stdin please."
		return
	#end if
	screen1 = None
	
	try:
		if use_screen:
			screen1 = init_curses()
		#end if
		if screen1:
			draw_buckets(screen1, {}, {})
		#end if
		read_packet_loop(infile, ip, screen1)
	except KeyboardInterrupt:
		pass
	except IOError, e:
		# This happens when we resize screen during infile.readline().
		print 'IOError', e.message
		pass
	except Exception, e:
		# If there's an error on the code, end_curses then print it.
		# Otherwise you'll end up with a broken term and an ugly error.
		if use_screen:
			end_curses(screen1)
		#end if
		raise
	#end try
	if use_screen:
		end_curses(screen1)
	#end if
	print "Exiting."
#end def main()

if __name__ == '__main__':
	main()
#end if
